-------------------------------------------------------------------------
-- LuaClass
--
-- This is a library to support your object oriented programming,
-- All are written in pure Lua 5.0.
-- Easy to use polymorphic objects, like Python, Ruby, and Smalltalk.
--
--
-- * Low cost message dispatching emulation.
--   LuaClass send message using `__index' field, built in Lua language.
--
-- * Complete late binding.
--   You can add and re-define class's methods after create subclasses.
--   Also `singleton method' ( instance's unique method ) available.
--
--
-- [ special members of objects ]
--
--   super :    LuaClass setup this method.
--              You can call a same name method of super class, like this,
--              self:super( arg1, arg2, ... )
--
--   init :     You can initialize your object in this method.
--              Called from YourClass:new() after instance creation.
--              init() receives all arguments from YourClass:new().
--
--   __foobar : Members that name begins with '__' (double underscore) is
--              defined by system. Don't touch it without necessity,
--              LuaClass depends on that these members exists.
--              Printing methods hide these members from you.
--
--
-- $Id: LuaClass.lua,v 1.8 2005/08/17 14:26:22 okayu Exp $
-------------------------------------------------------------------------
require "declare"
require "bind"
require "base/LuaClassImplement"
require "base/Object"
require "base/Class"
--~ require "base/Module"
require "base/BasicCollections"

begin_declare()



-- get list of sigleton methods which an object has.
function Object.__index:singletonMethods()
	local methods = Dictionary()
	table.foreach(
		self,
		function(k, v)
			if type(v) == "function"
				then methods:add(k, v)
			end
		end
	)
	return methods
end


-- get list of all methods which an object has.
function Object.__index:allMethods()
	local methods = Dictionary()
	local inherits = self:class():inherits():reverse()
	inherits:foreach(
		function(i, class)
			methods:addAll( class:definedMethods() )
		end
	)
	methods:addAll( self:singletonMethods() )
	return methods
end


-- get list of inherited classes.
function Class.__index:inherits()
	local array = Array()
	local class = self
	while class ~= nil do
		array:add( class )
		class = class:superclass()
	end
	return array
end



-- get list of methods defined by a class.
--
-- note:
--   `__index' field is not an object. It's a table of methods.
--   But there is a metatable of `__index', it is a super class of owner class.
--   LuaClass.isObject() mistakes `__index' as an object.
--   `__index' should not be handled as it is.
--   This function returns `__index' as Dictionary.
function Class.__index:definedMethods()
	return Dictionary( self.__index )
end;


-- get list of all classes.
function Class:classPool()
	local set = Set()
	set:addAll( self.mClassPool )
	return set
end


-- return string that inspects members of the object / table.
--
-- object : target object
-- maxExpand : limit for recursive expand members.
-- indent, index, str : internal use.
function LuaClass.inspectValue( object, maxExpand, indent, index, str )

	assert( type(object) == "table",  "inspectValue() : no value given.")
	str = str or ""
	indent = indent or 0
	maxExpand = maxExpand or 1
	local metaTable = getmetatable(object)
	local indentSpace = function(i) return string.rep("    ", i) end

	-- indent space.
	str = str..indentSpace(indent)

	-- table-index of this object.
	if index then
		str = str..index.." : "
	end

	-- digest string of this object.
	str = str..LuaClass.valueDigest(object)

	-- show item fold and return, if no more expand.
	if maxExpand < 0 then
		return str.." {...}\n"
	end

	-- expands members of this object.
	str = str.." {\n"
	indent = indent + 1

	-- show variables & methods which this object has.
	local fn = function(i, v)
		if string.sub(tostring(i), 1, 2)  == "__" then
			-- ignore special members which name begins `__'.
		elseif type(v) == "function" then
	        str = str..indentSpace(indent)..tostring(i).." : ("
	        	..LuaClass.valueDigest(v)..")\n"
		elseif LuaClass.isClass(v) then
			str = str..LuaClass.inspectValue(v, maxExpand - 2, indent, i)
		elseif LuaClass.isObject(v) then
			str = str..LuaClass.inspectValue(v, maxExpand - 2, indent, i)
		elseif type(v) == "table" then
			str = str..LuaClass.inspectValue(v, maxExpand - 1, indent, i)
		elseif type(v) == "string" then
	        str = str..indentSpace(indent)..tostring(i)..' : "'
	        	..LuaClass.toRawString(v)..'"\n'
		else
	        str = str..indentSpace(indent)..tostring(i)
				.." : "..tostring(v).."\n"
		end
	end

	table.foreach( object, fn )

	-- show methods which provided by the class of this object.
	if LuaClass.isObject(object) then
		str = str.."\n"..indentSpace(indent).."---- provided by `"
			..object:className().."' class ----".."\n"

		local fn = function(i, v)
			if string.sub(tostring(i), 1, 2)  == "__" then
				-- ignore special members which name begins `__'.
			elseif type(v) == "function" then
				str = str..indentSpace(indent)..tostring(i).." : ("
					..LuaClass.valueDigest(v)..")\n"
		    else
		    	-- ignore if not function.
		    end
		end

		table.foreach( LuaClass.getClass(object).__index, fn )
	end

	return str..indentSpace(indent - 1).."}\n"
end


-- inspect all object's members.
function Object.__index:inspect( maxExpand, indent, index, str )
	assert( LuaClass.isObject(self), "no self given." )
	return LuaClass.inspectValue( self, maxExpand, indent, index, str )
end



-------------------------------------------------------------------------
-- end
-------------------------------------------------------------------------
end_declare()


-------------------------------------------------------------------------
-- Copyright (C) 2005 Okayu Punch.
--
-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation files
-- (the "Software"), to deal in the Software without restriction,
-- including without limitation the rights to use, copy, modify,
-- merge, publish, distribute, sublicense, and/or sell copies of the
-- Software, and to permit persons to whom the Software is furnished
-- to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
-- ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
-- CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
-- WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-------------------------------------------------------------------------
