-------------------------------------------------------------------------
-- xUnit porting.
--
-- $Id: LuaUnit.lua,v 1.4 2005/05/31 13:38:40 okayu Exp $
-------------------------------------------------------------------------
require "LuaClass"

-- Class name declarations.
begin_declare()
declare("TestResult")
declare("TestCase")
declare("TestSuite")

declare("assertEquals")
declare("assertNotEquals")
declare("assertSame")
declare("assertNotSame")
declare("assertTrue")
declare("assertFalse")


-------------------------------------------------------------------------
-- assertion functions.
-------------------------------------------------------------------------
local _failAssertion
local _failTag = "!TestCaseFailAssertTag!"
local _failTagStartPos = -string.len(_failTag)


-- assertEquals( expected, actual )
-- assertEquals( message, expected, actual )
function assertEquals( message, expected, actual )
	if actual == nil then
		message, expected, actual = nil, message, expected
	end
	if expected == actual then return end
	_failAssertion( message, expected, actual, "expected" )
end

-- assertNotEquals( expected, actual )
-- assertNotEquals( message, expected, actual )
function assertNotEquals( message, expected, actual )
	if actual == nil then
		message, expected, actual = nil, message, expected
	end
	if expected ~= actual then return end
	_failAssertion( message, expected, nil, "expected not equals" )
end

-- assertSame( expected, actual )
-- assertSame( message, expected, actual )
function assertSame( message, expected, actual )
	if actual == nil then
		message, expected, actual = nil, message, expected
	end
	if rawequal(expected, actual) then return end
	_failAssertion( message, expected, actual, "expected same" )
end

-- assertNotSame( expected, actual )
-- assertNotSame( message, expected, actual )
function assertNotSame( message, expected, actual )
	if actual == nil then
		message, expected, actual = nil, message, expected
	end
	if rawequal(expected, actual) == false then return end
	_failAssertion( message, expected, nil, "expected not same" )
end

-- assertTrue( expected )
-- assertTrue( message, expected )
function assertTrue( message, expected )
	if expected == nil then message, expected = nil, message end
	if expected == true then return end
	_failAssertion( message, nil, expected, "expected true" )
end

-- assertFalse( expected )
-- assertFalse( message, expected )
function assertFalse( message, expected )
	if expected == nil then message, expected = nil, message end
	if expected == false then return end
	_failAssertion( message, nil, expected, "expected false" )
end


-- create message string and raise error.
-- ( local function )
function _failAssertion( message, expected, actual, assertType )
	assertType = assertType or ""
	local str

	str = "AssertFailed.\n"

	if message ~= nil then
		str = str..tostring(message).." "
	end

	str = str..assertType

	if expected ~= nil then
		str = str..':<'..LuaClass.valueDigest(expected)..'>'
	end

	if actual ~= nil then
		str = str..', but was:<'..LuaClass.valueDigest(actual)..'>'
	end

	str = str.._failTag

	error( str, 3 )		-- raise error.
end



-------------------------------------------------------------------------
-- TestResult Class
-------------------------------------------------------------------------

TestResult = { "TestResult"; }

	-- instance vaiables : mRunCount, mPassCount, mFailures, mErrors

	function TestResult:init()
		self.mRunCount = 0
		self.mPassCount = 0
		self.mFailures = Array()
		self.mErrors = Array()
	end


	function TestResult:testStarted()
		self.mRunCount = self.mRunCount + 1
	end

	function TestResult:testPassed()
		self.mPassCount = self.mPassCount + 1
	end

	function TestResult:addError( err )
		self.mErrors:add( err )
	end

	function TestResult:addFailure( failure )
		 self.mFailures:add( failure )
	end


	function TestResult:runCount()
		return self.mRunCount
	end

	function TestResult:passCount()
		return self.mPassCount
	end

	function TestResult:errorCount()
		return self.mErrors:size()
	end

	function TestResult:failureCount()
		return self.mFailures:size()
	end


	function TestResult:errors()
		return self.mErrors
	end

	function TestResult:failures()
		return self.mFailures
	end


	function TestResult:summary()
		return string.format("%d runs, %d errors, %d failures",
			self:runCount(), self:errorCount(), self:failureCount())
	end


Class:new ( TestResult )



-------------------------------------------------------------------------
-- TestCase Class
-------------------------------------------------------------------------

TestCase = { "TestCase"; }

	-- instance variables : mName

	function TestCase:init( name )
		self.mName = name
	end

	function TestCase:setUp()
		-- do nothing
	end

	function TestCase:tearDown()
		-- do nothing
	end

	function TestCase:run( result )
		result = result or TestResult()

		local phase
		local testMethod = self:getMethod(self.mName)
		if testMethod == nil then
			error( "`"..self.mName.."' method is not implemented." )
		end

		-- error handling closure
		local onError = function(err)
			local str
			if string.sub(err, _failTagStartPos, -1) == _failTag then
				-- Assertion Failed.
				err = string.sub(err, 1, _failTagStartPos - 1)
				result:addFailure(err)
			else
				-- Error
				str = "\n<"..self:className()..' "'..self.mName..'">\n'
					.."in "..phase.." phase, raise an Error.\n"
					..err
					..debug.traceback(err)
				result:addError( str )
			end
		end

		-- closure to setUp
		local doSetUp = self:bind( self.setUp )

		-- closure to run
		local doRun = self:bind( testMethod )

		-- closure to tearDown
		local doTearDown = self:bind( self.tearDown )

		-- run test.
		local isSetupOK = false
		local isRunOK = false
		result:testStarted()

		phase = "setUp"
		isSetupOK = xpcall(doSetUp, onError)

		if isSetupOK then
			phase = "run"
			isRunOK = xpcall(doRun, onError)
		end

		phase = "tearDown"
		xpcall(doTearDown, onError)

		if isRunOK then
			result:testPassed()
		end

		return result
	end


Class:new ( TestCase )


-------------------------------------------------------------------------
-- TestSuite
-------------------------------------------------------------------------

TestSuite = { "TestSuite"; }

	-- instance variables : mName, mTests

	-- TestSuite:init()
	-- TestSuite:init( name )
	-- TestSuite:init( testClass, )
	-- TestSuite:init( testClass, name )
	function TestSuite:init( testClass, name )
		if type(testClass) == "string" then
			testClass, name = nil, testClass
		end
		self.mName = name
		self.mTests = Array()

		if testClass then
			self:addTestSuite( testClass )
			if self.mName == nil then
				self.mName = testClass:instanceName()
			end
		end
	end

	function TestSuite:add( test )
		self.mTests:add( test )
	end

	-- add all methods begin with `test' of the test class.
	function TestSuite:addTestSuite( testClass )
		assert( LuaClass.isKindOfClass(testClass, TestCase),
			"TestCase class expected." )

		-- collect test methods of all inherited classes.
		local testNameSet = Set()
		local inherits = testClass:inherits()
		local addTest = function(name, method)
			if self:isTestMethod(name) then
				testNameSet:add(name)
			end
		end
		local addAllTests = function(i, class)
			class:definedMethods():foreach( addTest )
		end
		inherits:foreach( addAllTests )

		-- add methods.
		testNameSet:foreach(
			function(i, name)
				self.mTests:add( testClass:new(name) )
			end
		)
	end

	function TestSuite:run( result )
		result = result or TestResult()
		self.mTests:foreach(
			function(i, v) v:run(result) end
		)
	end


	function TestSuite:isTestMethod( methodName )
		return string.sub(methodName, 1, 4) == "test"
	end

	function TestSuite:testAt( index )
		return self.mTests.at(index)
	end

	function TestSuite:setName( name )
		self.mName = name
	end

	function TestSuite:getName()
		return self.mName or "no name"
	end


Class:new ( TestSuite )



-------------------------------------------------------------------------
-- end of file
-------------------------------------------------------------------------
end_declare()