-------------------------------------------------------------------------
-- Basic Collection classes
--
-- $Id: BasicCollections.lua,v 1.9 2005/07/07 17:17:22 okayu Exp $
-------------------------------------------------------------------------

-- Class name declarations.
begin_declare()

declare("Collection")
declare("OrderedCollection")
declare("MappedCollection")

declare("RawArray")
declare("Array")
declare("Dictionary")
declare("Bag")
declare("Set")
declare("Stack")
declare("WeakArray")
declare("WeakSet")
declare("WeakDictionary")


-------------------------------------------------------------------------
-- Collection class (Abstract)
-------------------------------------------------------------------------
Collection = { "Collection"; class = {}; }

	-- instance variables :
	
	function Collection:init( _table )
		if _table then self:initByTable( _table ) end
	end
	
	-- initialize by table.
	function Collection:initByTable( _table )
		self:addAll( _table )
	end

	
	-- return the number of elements.
	Collection.size = Object.subclassResponsibility
	
	-- do for all elements.
	Collection.foreach = Object.subclassResponsibility

	--  add
	Collection.add	 = Object.subclassResponsibility
	
	--  add all elements of collection or table.
	function Collection:addAll( aCollection )
		if LuaClass.isKindOfObject(aCollection, Collection) then
			aCollection:foreach( function(i, v) self:add(v) end )
		else
			table.foreach( aCollection,
				function(i, v) self:add(v) end
			)
		end
	end;

	--  remove
	Collection.remove	 = Object.subclassResponsibility
	
	-- has some value?
	Collection.hasValue = Object.subclassResponsibility
	

	-- is empty?
	function Collection:isEmpty()
		return self:size() == 0
	end
	
	-- create string to represent no key collection info.
	function Collection:tostring()
		local str = "{ "
		local firstTime = true
		-- enumurate all elements.
		self:foreach(
			function(i, v)
				if firstTime then firstTime = false
				else str = str..", "
				end
				str = str..LuaClass.valueDigest(v)
			end
		)
		return str.." }"
	end
		
	-- class method.
	Collection.class.new = Object.notSupported

	
Class:new( Collection )


-------------------------------------------------------------------------
-- OrderedCollection class (Abstract)
-------------------------------------------------------------------------
OrderedCollection = { "OrderedCollection", "Collection"; class = {}; }
	
	-- instance variables :
	
	function OrderedCollection:init( _table )
		self:super(_table)
	end
	
	-- initialize by table.
	function OrderedCollection:initByTable( _table )
		-- initialize by table.
		assert( type(_table) == "table", "string argument expected." )
		if LuaClass.isKindOfObject(_table, OrderedCollection) then
			-- copy all elements from an ordered collection.
			_table:foreach(
				function(index, value) self:insert(index, value) end
			)
		else
			-- copy from not an ordered collection.
			self:addAll( _table )
		end
	end

	--  at
	OrderedCollection.at = Object.subclassResponsibility
	
	-- first element
	function OrderedCollection:first()
		return self:at(1)
	end
	
	-- last element
	function OrderedCollection:last()
		return self:at(size())
	end

	
	-- push to last
	function OrderedCollection:pushBack( value )
		return self:add(value)
	end
	
	-- push to first
	function OrderedCollection:pushFront( value )
		return self:insert(1, value)
	end

	-- pop from last
	function OrderedCollection:popBack()
		return self:remove()
	end
	
	-- pop from first
	function OrderedCollection:popFront()
		return self:remove(1)
	end

	OrderedCollection.push = OrderedCollection.pushBack
	OrderedCollection.pop = OrderedCollection.popBack

	
	-- append a collection at the tail.
	OrderedCollection.append = Collection.addAll

	
	-- has some value?
	function OrderedCollection:hasValue( value )
		return self:find(value) ~= nil
	end

	-- find value. return index.
	function OrderedCollection:find( value )
		return self:foreach(
			function(i, v)
				if v == value then return i end
			end
		)
	end
	
	-- create a reverse ordered array.
	function OrderedCollection:reverse()
		local array = RawArray()
		local i
		for i = self:size(), 1, -1 do
			array:add( self:at(i) )
		end
		return array
	end

	-- class method.
	OrderedCollection.class.new = Object.notSupported
	
Class:new( OrderedCollection )


-------------------------------------------------------------------------
-- MappedCollection class (Abstract)
-------------------------------------------------------------------------
MappedCollection = { "MappedCollection", "Collection"; class = {}; }
	
	-- instance variables :
	
	function MappedCollection:init( _table )
		self:super(_table)
	end

	
	-- at
	MappedCollection.at = Object.subclassResponsibility
	
	
	-- has some value?
	function MappedCollection:hasValue( value )
		return self:find(value) ~= nil
	end

	-- has some key?
	function MappedCollection:hasKey( key )
		return self:at(key) ~= nil
	end
	
	-- find value. return key.
	function MappedCollection:find( value )
		return self:foreach(
			function(i, v)
				if v == value then return i end
			end
		)
	end
	
	-- create string to represent keyed collection info.
	function MappedCollection:tostring()
		local str = "{ "
		local firstTime = true
		-- enumurate all elements.
		self:foreach(
			function(key, value)
				if firstTime then firstTime = false
				else str = str..", "
				end
				str = str..key.." > "..LuaClass.valueDigest(value)
			end
		)
		return str.." }"
	end
	
	-- append a collection.
	function MappedCollection:addAll( aCollection )
		if LuaClass.isKindOfObject(aCollection, Collection) then
			assert(
				aCollection:isKindOf(OrderedCollection) or
				aCollection:isKindOf(MappedCollection),
				"MappedCollection:addAll()  argument collection is not keyed,"
				.." and not ordered."
			)
			aCollection:foreach(
				function(k, v) self:add(k, v) end
			)
		
		else
			table.foreach(
				aCollection,
				function(k, v) self:add(k, v) end
			)
		end
	end
	
	-- class method.
	MappedCollection.class.new = Object.notSupported
	
Class:new( MappedCollection )


-------------------------------------------------------------------------
-- RawArray class
--
-- brief usage :
--   array = RawArray { "A", 120, "C", "D", "abcde", Object }
--   print( array[1] )
--
-- note :
--   the index of array begins from `1'.
--
-- note :
--   RawArray behaves same as array of Lua table.
--   It will not work well that assigned by out-of-range indexed operator.
--   So, add() and insert() are recommened. for example:
--
--     array = RawArray { "1st", "2nd" }
--     print( array:size() )         -- 2
--
--     -- assign to out of range. there is no error,
--     -- but it is seem to be not added to array.
--     array[3] = "3rd"              
--     print( array:size() )         -- 2
--     print( array[3] )             -- 3rd
--     print( array )                -- { "1st", "2nd" }
--
--     -- you may set the size of array yourself.
--     array:setSize(3)
--     print( array:size() )         -- 3
--     print( array[3] )             -- 3rd
--     print( array )                -- { "1st", "2nd", "3rd" }
--
--     -- add() works well.
--     array:add("4th")
--     print( array:size() )         -- 4
--     print( array[4] )             -- 4th
--     print( array )                -- { "1st", "2nd", "3rd", "4th" }
--
--     -- insert() works well too.
--     array:insert(100, "100th")
--     print( array:size() )         -- 100
--     print( array[100] )           -- 100th
--     print( array )                -- { "1st", "2nd", "3rd", "4th",
--                                   --   nil, nil, ....,..........., "100th" }
-------------------------------------------------------------------------
RawArray = { "RawArray", "OrderedCollection"; }

	-- instance variables :
	
	function RawArray:init( _table )
		self:setSize(0)
		self:super(_table)
	end
	
	function RawArray:at( index )
		if type(index) ~= "number" then return nil end
		return self[index]
	end

	function RawArray:add( value )
		return table.insert(self, value)
	end
	
	function RawArray:remove( index )
		return table.remove(self, index)
	end
	
	function RawArray:setSize( size )
		return table.setn(self, size)
	end
	
	RawArray.foreach = table.foreachi
	
	RawArray.size = table.getn
	
	RawArray.insert = table.insert

Class:new( RawArray )


-------------------------------------------------------------------------
-- Array class
--
-- brief usage :
--   array = Array { "A", 120, "C", "D", "abcde", Object }
--   print( array:at(1) )
--   array:insert( 2, "B" )
--   array:remove( 3 )
--   print( array )
--
-- note :
--   the index of array begins from `1'.
-------------------------------------------------------------------------
Array = { "Array", "OrderedCollection"; }

	-- instance variables : mData
	
	function Array:init( _table )
		self.mData = {}
		self:super(_table)
	end
	
	function Array:add( value )
		return table.insert(self.mData, value)
	end
	
	function Array:remove( index )
		return table.remove(self.mData, index)
	end
	
	function Array:insert( index, value )
		if value == nil then
			return table.insert(self.mData, index)
		end
		assert( type(index) == "number", "numebr type expected." )
		return table.insert(self.mData, index, value)
	end
	
	function Array:at( index )
		return self.mData[index]
	end
	
	function Array:size()
		return table.getn(self.mData)
	end
	
	function Array:foreach( fn )
		return table.foreachi(self.mData, fn)
	end

	function Array:sort( comp )
		table.sort( self.mData, comp )
	end

Class:new( Array )



-------------------------------------------------------------------------
-- MDArray class
-- Multi-Dimensional Array.
--
-- This class does not have real deimensions.
-- Internaly, it's implemented as a dictionary collection.
-- MDArray translations indexing, for example:
--  array.at(100, 5, 60)  ->  array.at("100,005,060")
--
-- brief usage :
--   array = MDArray( 10, 10, 10 )
--   print( array:at(1, 1, 1) )    -- 0
--
-- MDArray( size1, [size2, ...] )
--   initialize by size of each dimension.
--   all values are nil initially.
--
-- MDArray:fill( value )
-- MDArray:fill( value, range1, range2, ... )
--   Whole elements are initialized by initValue.
-------------------------------------------------------------------------



-------------------------------------------------------------------------
-- Dictionary class
--
-- No order.
-- Similar to `association list' of Perl, or `map' of C++ STL.
-- Basically, same as `table' of Lua.
--
-- brief usage :
--   dict = Dictionary { firstItem = "A", secondItem = 120 }
--   print( dict:at("firstItem") )
--   dict:add( "secondItem", 2 )
--   dict:remove( "firstItem" )
--   print( dict )
--
-- note :
--   the index of array begins from `1'.
-------------------------------------------------------------------------
Dictionary = { "Dictionary", "MappedCollection"; }

	-- instance variables : mData, mCount
	
	function Dictionary:init( _table )
		self.mData = {}
		self.mCount = 0
		self:super(_table)
	end
	
	function Dictionary:add( key, value )
		self.mCount = self.mCount + 1
		self.mData[key] = value
	end

	function Dictionary:remove( key )
		local tmp = self.mData[key]
		if tmp == nil then return nil end
		self.mCount = self.mCount + 1
		self.mData[key] = nil
		return tmp
	end
	
	function Dictionary:at( key )
		return self.mData[key]
	end
	
	function Dictionary:size()
		return self.mCount
	end
	
	function Dictionary:foreach( fn )
		return table.foreach(self.mData, fn)
	end
	
Class:new( Dictionary )


-------------------------------------------------------------------------
-- Bag class
--
-- No order. Duplicates are allowed.
-- similar to `multi-set' of C++ STL.
--
-- brief usage :
--   bag = Bag { "A", 120, "C", "D", 120, "abcde", Object }
--   print( bag )    -- { "A", Object class, "C", 120, 120, "D", "abcde" }
-------------------------------------------------------------------------
Bag = { "Bag", "Collection"; }

	-- instance variables : mData, mCount
	
	function Bag:init( _table )
		self.mData = {}
		self.mCount = 0
		self:super(_table)
	end

	
	function Bag:size()
		return self.mCount
	end
	
	function Bag:add( value )
		if self.mData[value] then
			self.mData[value] = self.mData[value] + 1
		else
			self.mData[value] = 1
		end
		self.mCount = self.mCount + 1
	end
	
	function Bag:remove( value )
		local tmp = self.mData[value]
		if tmp == nil then return nil end
		if tmp == 1 then
			self.mData[value] = nil
		else
			self.mData[value] = tmp - 1
		end
		self.mCount = self.mCount - 1
		return tmp
	end
	
	-- note :
	--  Bag has no index.
	--  fn (i, v) ... `i' always receives nil.
	function Bag:foreach( fn )
		return table.foreach(
			self.mData,
			function(value, count)
				local n
				for n = 1, count do
					local ret = fn(nil, value)
					if ret then return ret end
				end
			end
		)
	end
	
	function Bag:hasValue( value )
		return self.mData[value] ~= nil
	end
	
	function Bag:countOf( value )
		return self.mData[value] or 0
	end

Class:new( Bag )



-------------------------------------------------------------------------
-- Set class
--
-- No order. Duplicates are not allowed.
--
-- brief usage :
--   set = Set { "A", 120, "C", "D", 120, "abcde", Object }
--   print( set )   -- { "A", 120, "C", "D", "abcde", Object class }
-------------------------------------------------------------------------
Set = { "Set", "Collection"; }
	
	-- instance variables : mData

	function Set:init( _table )
		self.mData = {}
		self:super(_table)
	end
	
	
	function Set:size()
		return table.getn(self.mData)
	end
	
	function Set:add( value )
		if not self.mData[value] then
			self.mData[value] = 1
			table.setn(self.mData, self:size() + 1)
		end
	end
	
	function Set:remove( value )
		local tmp = self.mData[value]
		if tmp == nil then return nil end
		self.mData[value] = nil
		table.setn(self.mData, self:size() - 1)
		return tmp
	end
	
	-- note :
	--  Set has no index.
	--  fn (i, v) ... `i' always receives nil.
	function Set:foreach( fn )
		return table.foreach(
			self.mData,
			function(value, dummy) return fn(nil, value) end
		)
	end
	
	function Set:hasValue( value )
		return self.mData[value] ~= nil
	end
	

Class:new( Set )



-------------------------------------------------------------------------
-- Stack class
-------------------------------------------------------------------------
Stack = { "Stack", "RawArray"; }

	-- instance variables :

	function Stack:initialize( _table )
		self:super( _table )
	end

	
	function Stack:push( value )
		return self:insert( value )
	end
	
	function Stack:pop()
		return self:remove()
	end

Class:new( Stack )



-------------------------------------------------------------------------
-- WeakArray class
--
-- Array with weak refference of elements.
-- Garbage collector can free elements in a WeakArray.
-- The removed elements are set nil.
-------------------------------------------------------------------------
WeakArray = { "WeakArray", "Array"; }

	-- instance variables :

	function WeakArray:init( _table )
		self:super()
		setmetatable(self.mData, LuaClass.weakValueTable)
		if _table then self:initByTable( _table ) end
	end

Class:new( WeakArray )



-------------------------------------------------------------------------
-- WeakSet class
--
-- Set with weak refference of elements.
-- Garbage collector can free elements in a WeakSet.
-------------------------------------------------------------------------
WeakSet = { "WeakSet", "Set"; }	

	-- instance variables :

	function WeakSet:init( _table )
		self:super()
		setmetatable(self.mData, LuaClass.weakKeyTable)
		if _table then self:initByTable(_table) end
	end
	
	
	function WeakSet:size()
		local count = 0
		self:foreach( function() count = count + 1 end )
		return count
	end
	
Class:new( WeakSet )


-------------------------------------------------------------------------
-- WeakDictionary class
--
-- Dictionary with weak refference of elements.
-- Garbage collector can free elements in a WeakDictionary.
-------------------------------------------------------------------------
WeakDictionary = { "WeakDictionary", "Dictionary"; }

	-- instance variables :
	
	function WeakDictionary:init( _table )
		self:super()
		setmetatable(self.mData, LuaClass.weakValueTable)
		if _table then self:initByTable( _table ) end
	end

	function WeakDictionary:size()
		local count = 0
		self:foreach( function() count = count + 1 end )
		return count
	end

Class:new( WeakDictionary )



-------------------------------------------------------------------------
-- end of file
-------------------------------------------------------------------------
end_declare()