-------------------------------------------------------------------------
-- Class class
--
-- $Id: Class.lua,v 1.9 2005/07/02 15:02:04 okayu Exp $
-------------------------------------------------------------------------
-- Class name declarations.
begin_declare()
declare("Class")


-------------------------------------------------------------------------
--  `Class' class definition
--
--  This is the class of all classes.
--  The `Class' class works as a meta class.
--
--
--  There are some special class-members for LuaClass, as follows,
--
--  variables :
--    __classname__, __superclass__, __subclasses__,
--    __index, __addindex, __metatable, __mode
--
--  methods :
--    __add, __sub, __mul, __div, __pow, __unm
--    __concat, __eq, __lt, __le, __call, __tostring
--
--  And all members which name begins with `__' are special.
--  Don't modify these members,
--  or the Messaging Mechanism will be confused and maybe crashed.
--  Object:inspect() method ignores these special members.
-------------------------------------------------------------------------

Class = {
	-- class information.
	__classname__ = "Class";	-- class name.
	__superclass__ = Object;	-- super class.
	__subclasses__ = {};

	-- operators supported by `Class'.
	__eq = LuaClass.__eq;
	__tostring = LuaClass.__tostring;
	__call = LuaClass.__call;

	mClassPool = {};
}


-------------------------------------------------------------------------
-- index of all class objects.
-------------------------------------------------------------------------
Class.__index = {

	-------------------- instance creation --------------------

	new = LuaClass.newObject;


	-------------------- testing --------------------

	-- is subclass of some class?
	isSubclassOf = LuaClass.isKindOfClass;


	-------------------- accessing --------------------

	-- return instance's name.
	instanceName = function(self) return self.__classname__ end;

	-- return class object.
	class = function() return Class end;

	-- return super class object.
	superclass = function(self) return self.__superclass__ end;

	-- return subclass table.
	subclasses = function(self) return self.__subclasses__ end;

	-- get list of inherited classes. (implemented in LuaClass.lua)
	inherits = nil;


	-------------------- message handling --------------------

	-- operator() applied as call a function. it is an ailias to new().
	call = function(self, ...) return self:new(unpack(arg)) end;


	-------------------- method creation --------------------

	-- add method in a class.
	addMethod = function( self, methodName, fn )
		self.__index[methodName] =
			LuaClass.addMethod( self, methodName, fn )
	end;

	-- get list of methods defined by a class. (implemented in LuaClass.lua)
	definedMethods = nil;

}


-------------------------------------------------------------------------
-- setup the Object and Class.
-------------------------------------------------------------------------
setmetatable( Object, Class )
setmetatable( Class, Class )
setmetatable( Class.__index, Object )
setmetatable( Object.__subclasses__, LuaClass.weakValueTable )
setmetatable( Class.__subclasses__, LuaClass.weakValueTable )

setmetatable( Class.mClassPool, LuaClass.weakValueTable )
Class.mClassPool[ Object.__classname__ ] = Object
Class.mClassPool[ Class.__classname__ ] = Class
table.insert( Object.__subclasses__, Class )


-------------------------------------------------------------------------
-- Class:new()
--
-- create new class.
--
-- Class:new {
--   newClassName
--   [, superClassName]
--   [, moduleName] [, moduleName]...
--
--   method_definitions
--
--   class = {
--     classMethod definitions...
--   }
-- }
--
-------------------------------------------------------------------------
function Class:new( definition )

	assert( rawequal(self, Class),
		LuaClass.noSelfErrorString( "Class:new()", "Class", "new")
	)
	assert( type(definition) == "table",
	 	"Class:new() : need table argument.")

	local className = definition[1]
	local superName = definition[2] or "Object"
	definition[1] = nil
	definition[2] = nil

	local classMethods
	if definition["class"] ~= nil then
		classMethods = definition["class"]
		definition["class"] = nil
	end

	assert( type(className) == "string",
		"Class:new() : The class name must be string type." )
	assert( type(superName) == "string",
		"Class:new() : The super class name must be string type." )
	assert( className, "Class:new() : no class name given." )
	assert( Class.mClassPool[ superName ],
		"Class:new() : Super class `"..superName.."' not exists." )
	assert( not Class.mClassPool[ className ],
		"Class:new() : The `"..className.."' class already exists.")

	-- setup class members
	local newClass = {}
	newClass.__classname__ = className
	newClass.__superclass__ = Class.mClassPool[ superName ]
	newClass.__subclasses__ = {}
	setmetatable( newClass.__subclasses__, LuaClass.weakValueTable )
	newClass.__index = {}
	newClass.__eq = LuaClass.__eq
	newClass.__tostring = LuaClass.__tostring

	-- setup inheritance.
	setmetatable( newClass, Class )
	setmetatable( newClass.__index, newClass.__superclass__ )
	table.insert( newClass:superclass():subclasses(), newClass )

	-- setup methods.
	LuaClass.initMethods( newClass, definition )

	-- setup class methods.
	if classMethods then
		LuaClass.initSingletonMethods( newClass, classMethods )
	end

	-- setup the environment.
	Class.mClassPool[ className ] = newClass
	declare( className )
	_G[className] = newClass

	-- special setup for `Class' subclass.
	if newClass:isSubclassOf(Class) then
		LuaClass.newClassSubclass( newClass )
	end

	-- initialize new class.
	newClass:init()

	return newClass

end



-------------------------------------------------------------------------
-- end of file
-------------------------------------------------------------------------
end_declare()