-------------------------------------------------------------------------
-- LuaClass implementations
--
-- $Id: LuaClassImplement.lua,v 1.7 2005/07/14 22:20:00 okayu Exp $
-------------------------------------------------------------------------
-- Table name declarations.
begin_declare()
declare("LuaClass")


-------------------------------------------------------------------------
--  LuaClass table
--
--  This table contains utilities, like a namespace.
-------------------------------------------------------------------------

LuaClass = {

	-- for weak refference.
	weakValueTable = { __mode = "v"; };
	weakKeyTable = { __mode = "k"; };
	
}


-- copy table.
function LuaClass.copyTable( src )
	local dst = {}
	table.foreach( src, function(i) dst[i] = src[i] end )
	return dst
end


-- create new class object.
function LuaClass.newObject( aClass, ... )
	assert( aClass,
		aClass:instanceName()..":new() : No self object given.\n"
		.." Don't you use "..aClass:instanceName()..".new instead of "
		..aClass:instanceName()..":new ?")

	local newObject = {}
	setmetatable( newObject, aClass )

	-- call object:init() method.
	newObject:init( unpack(arg) )
	return newObject
end


-- When `Class' subclass is created, Class:new() call this function.
function LuaClass.newClassSubclass( aClass )
	local oldFunc = LuaClass.isClass
	LuaClass.isClass = function( var )
		return oldFunc(var) or rawequal(LuaClass.getClass(var), aClass)
	end
end


-- return class of instance.
LuaClass.getClass = getmetatable


-- is var a instance of `Class' class?
function LuaClass.isClass( var )
	return rawequal(LuaClass.getClass(var), Class)
end


-- is var a instance of class inherit `Object'?
function LuaClass.isObject( var )
	return LuaClass.isClass( LuaClass.getClass(var) )
end


-- is aClass a instance of expectedClass? (not check inheritance)
function LuaClass.isInstanceOfClass( aClass, expectedClass )
	if rawequal(aClass, expectedClass) then
		return true
	end
	return false
end


-- is aClass a class inherit baseClass?
function LuaClass.isKindOfClass( aClass, baseClass )
	if aClass == nil then
		return false
	end
	if not LuaClass.isClass(aClass) then
		return false
	end
	if rawequal(aClass, baseClass) then
		return true
	end
	return LuaClass.isKindOfClass( aClass:superclass(), baseClass )
end


-- is object a instance of class inherit someClass?
function LuaClass.isKindOfObject( object, someClass )
	if object == nil then return false end
	return LuaClass.isKindOfClass( LuaClass.getClass(object), someClass )
end


-- not supported method.
function LuaClass.notSupported()
	error( "This method is not supported." )
end


-- subclass must define this method.
function LuaClass.subclassResponsibility()
	error( "Subclass must define this method." )
end


-- super() cannot pass this method.
function LuaClass.superNotDefined()
	error( "super() cannot pass this method." )
end


-- create accessing methods for private value.
-- usage :
--   getter, setter = LuaClass.accessor( value )
function LuaClass.accessor( value )
	return
		function() return value end,
		function(v) value = v end
end



-- operators for re-definition
function LuaClass.__add( op1, op2 )
	return op1:add(op2)
end

function LuaClass.__sub( op1, op2 )
	return op1:subtract(op2)
end

function LuaClass.__mul( op1, op2 )
	return op1:multiply(op2)
end

function LuaClass.__div( op1, op2 )
	return op1:divine(op2)
end

function LuaClass.__pow( op1, op2 )
	return op1:expOfTen(op2)
end

function LuaClass.__unm( op1 )
	return op1:unaryMinus()
end

function LuaClass.__concat( op1, op2 )
	return op1:concat(op2)
end

function LuaClass.__eq( op1, op2 )
	return op1:isEqual(op2)
end

function LuaClass.__lt( op1, op2 )
	return op1:isLargerThan(op2)
end

function LuaClass.__le( op1, op2 )
	return op1:isLargerEqual(op2)
end

function LuaClass.__call( op1, ... )
	return op1:call( unpack(arg) )
end

function LuaClass.__tostring( op1 )
	return op1:tostring()
end


-- initialize methods from table of methods.
function LuaClass.initMethods( object, methodList, isSingleton )
	isSingleton = isSingleton or false

	local fn = function (i, v)
		assert( type(v) == "function",
			"initMethods() : Illegal method detected!\n"
			.." `"..i.."' is not a function." )
		assert( type(i) == "string",
			"initMethods() : No named method detected : "..i )

		-- create method dispatcher.
		if isSingleton then
			object[i] = LuaClass.addSigletonMethod( object, i, v )
		else
			object.__index[i] = LuaClass.addMethod( object, i, v )
		end
	end

	table.foreach( methodList, fn )

end


-- initialize singleton methods from table of methods.
function LuaClass.initSingletonMethods( object, methodList )
	return LuaClass.initMethods( object, methodList, true )
end


-- return error message.
function LuaClass.noSelfErrorString( calleeName, objectName, methodName )
	return calleeName.." : No self object given.\n"
		.."Don't you call "
		..objectName.."."..methodName.."()"
		.." instead of "
		..objectName..":"..methodName.."() ?"
end


-- create Singleton Message dispatcher.
function LuaClass.addSigletonMethod( object, methodName, fn )
	assert( LuaClass.isObject(object),
		"addSigletonMethod() : object argument expected.")

	if fn == LuaClass.notSupported or
	   fn == LuaClass.subclassResponsibility then
		return fn	-- return well known error functions as it is,
					-- only for good looking in inspect()
	end

	return function(self, ...)
		assert(
			LuaClass.isKindOfObject(self, LuaClass.getClass(object)),
			LuaClass.noSelfErrorString( "<Message dispatcher>",
				object:instanceName(), methodName)
		)

		-- you can call self:super() to pass forward method defined in class.
		local saveSuper = self.super
		self.super = LuaClass.getClass(object).__index[methodName]
			or LuaClass.superNotDefined

		-- call your method.
		local ret = { fn(self, unpack(arg)) }

		self.super = saveSuper
		return unpack( ret )
	end
end


-- create Message dispatcher.
function LuaClass.addMethod( newClass, methodName, fn )

	if fn == LuaClass.notSupported or
	   fn == LuaClass.subclassResponsibility then
		return fn	-- return well known error functions as it is,
					-- only for good looking in inspect()
	end

	return function(self, ...)
		assert(
			LuaClass.isKindOfObject(self, newClass),
			LuaClass.noSelfErrorString( "<Message dispatcher>", 
				LuaClass.makeObjectName(newClass), methodName
			)
		)

		-- you can call self:super() to pass forward super class method.
		local saveSuper = self.super
		self.super = newClass:superclass().__index[methodName]
			or LuaClass.superNotDefined

		-- call your method.
		local ret = { fn(self, unpack(arg)) }

		self.super = saveSuper
		return unpack( ret )
		
		-- <<< note. these are not same. >>>
		-- local testVar = { 1, nil, 3 }
		-- print( "testVar = "..LuaClass.inspectValue( testVar ) )
		-- print( "{ unpack(testVar) } = "..LuaClass.inspectValue( { unpack(testVar) } ) )
	end
end


-- create a name like `aSomeClass' or `anObject'.
function LuaClass.makeObjectName( aClass )
	assert(LuaClass.isClass(aClass), "makeObjectName() : class expected.")
	local upperName = string.upper( aClass:instanceName() )
	if string.find("AEIOU", string.sub(upperName, 1, 1)) then
		return "an "..aClass:instanceName()
	end
	return "a "..aClass:instanceName()
end


-- return a instance name like `aSomeClass' or `anObject'.
function LuaClass.objectName( object )
	assert( LuaClass.isObject(object),  "objectName() : no object given.")
	return LuaClass.makeObjectName( LuaClass.getClass(object) )
end


-- return string as raw tostring() call.
function LuaClass.toRawString( op1 )
	if LuaClass.getClass(op1) then
		local escape = LuaClass.getClass(op1).__tostring
		LuaClass.getClass(op1).__tostring = nil
		local str = tostring(op1)
		LuaClass.getClass(op1).__tostring = escape
		return str
	end
	return tostring(op1)
end


-- return a digest string that represents a value.
function LuaClass.valueDigest( value )

	if type(value) == "table" then
		if LuaClass.isClass(value) then
			local superDigest = ""
			if value:superclass() then
				superDigest =
					" ("..value:superclass():instanceName().." subclass)"
			end
		    return value:instanceName().." class"..superDigest
		elseif LuaClass.isObject(value) then
		    return value:instanceName().." ("
		    	..LuaClass.toRawString(value)..")"
		end
	elseif rawequal(value, LuaClass.notSupported) then
		return "function : `notSupported'"
	elseif rawequal(value, LuaClass.subclassResponsibility) then
		return "function : `subclassResponsibility'"
	elseif rawequal(value, LuaClass.superNotDefined) then
		return "function : `superNotDefined'"
	elseif type(value) == "string" then
		return '"'..LuaClass.toRawString(value)..'"'
	end

	return LuaClass.toRawString(value)
end



-------------------------------------------------------------------------
-- end of file
-------------------------------------------------------------------------
end_declare()