-------------------------------------------------------------------------
-- Object class
--
-- $Id: Object.lua,v 1.7 2005/07/02 15:02:04 okayu Exp $
-------------------------------------------------------------------------
-- Class name declarations.
begin_declare()
declare("Object")


-------------------------------------------------------------------------
--  `Object' class definition
--
--  This is the root class of all objects.
-------------------------------------------------------------------------

Object = {
	-- class information.
	__classname__ = "Object";	-- class name.
	__superclass__ = nil;		-- `Object' is a root.
	__subclasses__ = {};		-- subclasses will be added.
}


-------------------------------------------------------------------------
-- root index of all objects.
-------------------------------------------------------------------------
Object.__index = {

	-------------------- initialize & release --------------------
	
	-- init() is called from className:new()
	-- return self if success, return nil if fail.
	init = function(self) return self end;
	
	-- This method is expected to release resources which object uses.
	release = LuaClass.subclassResponsibility;


	-------------------- accessing --------------------

	-- return instance's name.
	instanceName = LuaClass.objectName;

	-- return instance's class name.
	className = function(self)
		return LuaClass.getClass( self ):instanceName()
	end;

	-- return class object.
	class = LuaClass.getClass;
	
	-- return method that named methodName.
	getMethod = function( self, methodName )
		assert( type(methodName) == "string",
			"getMethod() : string argument expected." )
		if type(self[ methodName ]) ~= "function" then
			return nil
		end
		return self[ methodName ]
	end;

	-- define this after. (see LuaClass.lua)
	allMethods = nil;
	
	singletonMethods = nil;
		
	accessor = LuaClass.accessor;
	
	
	-------------------- copying --------------------

	-- return a copy of self object.
	clone = function(self)
		local obj = LuaClass.copyTable(self)
		setmetatable( obj, LuaClass.getClass(self) )
		return obj
	end;
	
	
	-------------------- testing --------------------

	-- is instance of expectedClass?
	isInstanceOf = function(self, expectedClass)
		return LuaClass.isInstanceOfClass(
			LuaClass.getClass( self ), expectedClass
		)
	end;

	-- is kind of baseClass?
	isKindOf = function(self, baseClass)
		return LuaClass.isKindOfClass(LuaClass.getClass(self), baseClass)
	end;

	
	-------------------- comparing --------------------

	-- is an object same as a target for identity?
	isSame = rawequal;

	-- is an object value same as target value for equality?
	-- this method override operator `=='.
	isEqual = rawequal;
	-- if you override `isSame' method, remove comment below.
--	isEqual = function(self, target) return self:isSame(target) end;

	
	-------------------- converting --------------------

	-- return string to represents an object.
	-- called automatically when necessary. see also Object.__tostring.
	tostring = LuaClass.valueDigest;

	-- same as tostring for convenience.
	toString = function(self) return self:tostring() end;
	
	-- return string that system represents an object.
	toRawString = LuaClass.toRawString;
	
	
	-------------------- printing --------------------


	
	-------------------- method creation --------------------

	-- add singlton method in this object.
	addSingletonMethod = function( self, methodName, fn )
		self[methodName] =
			LuaClass.addSigletonMethod( self, methodName, fn )
	end;

	
	-------------------- method binding --------------------

	-- bind function and arguments to this instance.
	bind = bindself;
	
	-- bind function and left arguments to this instance.
	bindLeft = bindselfLeft;
	
	-- bind function and right arguments to this instance.
	bindRight = bindselfRight;

	
	-------------------- error handling --------------------

	-- not supported method.
	notSupported = LuaClass.notSupported;

	-- subclass must define this method.
	subclassResponsibility = LuaClass.subclassResponsibility;
}

-------------------------------------------------------------------------
-- end of file
-------------------------------------------------------------------------
end_declare()